#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include "perfmon.h"

main(int argc, char *argv[])
{
    int fd;
    int rc;
    int i;
    int cpu = 0;
    char c;
    extern char *optarg;
    extern int optind;
    register unsigned long long a, b;
    unsigned long long min, max, total;
    int maxiter, miniter;
    register unsigned long x, y;
    register hrtime_t start, end;
    unsigned long long tmp;

    /* Parse command line arguments */
    while ((c = getopt(argc, argv, "p:")) != EOF)
	switch(c) {
	    /* -p cpu : binds this process to the indicated CPU */
	    case 'p':
		cpu = atoi(optarg);
		break;

	    default:
		break;
	}

    /* Try to bind us to the given CPU */
    rc = processor_bind(P_PID, P_MYID, cpu, NULL);
    if (rc == -1) {
	perror("processor_bind()");
	exit(1);
    }

    /* Try to open a channel to the perfmon driver */
    fd = open("/dev/perfmon", O_RDONLY);
    if (fd == -1) {
	perror("open(/dev/perfmon)");
	exit(1);
    }

    /* Set PCR to accumulate cycle count and number of instructions */
    tmp = PCR_S0_CYCLE_CNT | PCR_S1_INSTR_CNT | PCR_USER_TRACE | PCR_SYS_TRACE;
    rc = ioctl(fd, PERFMON_SETPCR, &tmp);
    if (rc < 0) {
	perror("ioctl(PERFMON_SETPCR)");
	exit(1);
    }

    /* Run Loop */
    printf("Running loop...");
    fflush(stdout);

    /* Try to figure out the number of machine cycles that gethrtime() takes */
    min = 100000000;
    max = total = 0;
#define LOOP_ITERS 1000000
    for (i = 0; i < LOOP_ITERS; i++) {
	clr_pic();
	cpu_sync();
	a = get_pic();

	start = gethrtime();

	cpu_sync();
	b = get_pic();

	x = extract_pic0(b) - extract_pic0(a);  /* Number of cycles */
	if (x > max) {
	    max = x;
	    maxiter = i;
	}

	if (x < min) {
	    min = x;
	    miniter = i;
	}

	total += x;
    }

    printf("Done.\n");

    printf("Results for gethrtime():\n");
    printf("\tMax number of cycles taken: %llu (iteration %d)\n", max, maxiter);
    printf("\tMin number of cycles taken: %llu (iteration %d)\n", min, miniter);
    printf("\tAvg number of cycles taken: %.2lf\n",
	(double)(total - min - max) / (LOOP_ITERS - 2));
}
